''' Input is a module which handles the genereation of input and submission files for VASP calculations. '''
# Impoting modules
import configparser as cp
import os
import socket
from .structure import Structure, get_extension


class InputFileGenerator(Structure):
    '''
    InputFileGenerator is a class which generates input files for VASP calculations.
    '''

    def __init__(self, filename: str, calculation: str):
        '''
        Initializes the class.
        '''
        extension = get_extension(filename)
        if extension == 'cif':
            cif = Structure.from_cif(filename)
            for atrr in cif.__dict__:
                setattr(self, atrr, getattr(cif, atrr))
            self.write_poscar('POSCAR')
        elif extension == 'vasp':
            poscar = Structure.from_poscar(filename)
            for atrr in poscar.__dict__:
                setattr(self, atrr, getattr(poscar, atrr))
        else:
            raise ValueError(f'Extension {extension} not found.')
        self.configuration = calculation

        hostname = socket.gethostname()
        if 'csf3' in hostname:
            self.pseudopotentials = '/opt/apps/apps/intel-19.1/vasp/5.4.4/pseudopotentials/potpaw_PBE.54'
            self.module = 'apps/intel-19.1/vasp/5.4.4'
        elif 'csf4' in hostname: 
            self.pseudopotentials = '/opt/software/RI/apps/VASP/5.4.4-iomkl-2020.02/pseudopotentials/potpaw_PBE.54'
            self.module = 'VASP/5.4.4-iomkl-2020.02'
        elif os.path.exists(
                os.path.join(
                    os.path.expanduser('~'),
                    '.vasp_suite_configs/host.ini')):
            config = cp.ConfigParser()
            config.read(os.path.join(os.path.expanduser('~'), '.vasp_suite_configs/host.ini'))
            try:
                self.pseudopotentials = os.path.join(
                        os.path.expanduser('~'), config['host']['pseudopotentials'])
                self.module = config['host']['module']
            except KeyError:
                print('''Create a host.ini file in ~/.vasp_suite_configs with the following format:
                [host]
                pseudopotentials = /path/to/pseudopotentials
                module = module_name''')
                raise KeyError(f'''No configuration found for {hostname}.''')
        else:
            raise Warning('Unable to generate POTCAR file. No configuration found for this host.')

    def generate_incar(self):
        config = cp.ConfigParser()
        input_file = os.path.join(os.path.expanduser('~'), '.vasp_suite_configs', f'''{self.configuration}.ini''')
        config.read(input_file)
        with open('INCAR', 'w') as f:
            for section in config.sections():
                f.write("! {}\n".format(section.upper()))
                for key in config[section]:
                    key = key.upper()
                    f.write(f'''{key} = {config[section][key]}\n''')

    def generate_potcar(self):
        pseudo = self.pseudopotentials
        potpaw = ['H', 'He', 'Li_sv', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na_pv', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K_sv', 'Ca_sv', 'Sc_sv', 'Ti_sv', 'V_sv', 'Cr_pv', 'Mn_pv', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga_d', 'Ge_d', 'As', 'Se', 'Br', 'Kr', 'Rb_sv', 'Sr_sv', 'Y_sv', 'Zr_sv', 'Nb_sv', 'Mo_sv', 'Tc_pv', 'Ru_pv', 'Rh_pv', 'Pd', 'Ag', 'Cd', 'In_d', 'Sn_d', 'Sb', 'Te', 'I', 'Xe', 'Cs_sv', 'Ba_sv', 'La', 'Ce_3', 'Nd_3', 'Pm_3', 'Sm_3', 'Eu_2', 'Gd_3', 'Tb_3', 'Dy_3', 'Ho_3', 'Er_3', 'Tm_3', 'Yb_2', 'Lu_3', 'Hf_pv', 'Ta_pv', 'W_sv', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl_d', 'Pb_d', 'Bi_d', 'Po_d', 'At', 'Rn', 'Fr_sv', 'Ra_sv', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm']
        atoms = self.atoms

        files = [x.split('_') for x in potpaw if x.split('_')[0] in atoms]
        for i in range(len(atoms)):
            for j in range(len(files)):
                if files[j][0] == atoms[i]:
                    files[j].append(i)
        files.sort(key=lambda x: x[-1])
        files = [x[:-1] for x in files]
        files = ['_'.join(x) for x in files]
        cwd = os.getcwd()

        with open('POTCAR', 'w') as f:
            os.system(f'''module load vasp''')
            os.chdir(pseudo)
            for file in files:
                os.chdir(str(file))
                with open('POTCAR', 'r') as g:
                    f.write(g.read())
                os.chdir(pseudo)
            os.chdir(cwd)



import xyz_py
import numpy as np
import matplotlib.pyplot as plt
from ase.data import covalent_radii, atomic_numbers, atomic_masses
from scipy.spatial.distance import cdist
import sys

# from .structure import Structure


class Node:

    def __init__(self, coordinate: np.ndarray, symbol: str, radius: float, idx: int, mass: float, scaler=1.1):
        self.coordinate = coordinate
        self.symbol = symbol
        self.radius = radius
        self.idx = idx
        self.mass = mass
        self.scaler = scaler

    def __sub__(self, other):
        return np.linalg.norm(self.coordinate - other.coordinate)

    def is_connected(self, other) -> bool:
        return self - other <= self.scaler * (self.radius + other.radius)

    @property
    def xy_project(self) -> np.ndarray:
        return self.coordinate @ np.array([
            [1, 0],
            [0, 1],
            [0, 0]
        ])


class MolGraph:
    def __init__(self, scaler=1.1):
        self.nodes = []
        self.edges = {}
        self.crystal = False
        self.scaler = scaler

    def add_node(self, node):
        if type(node) is Node:
            self.nodes.append(node)
        elif type(node) is list:
            for n in node:
                self.nodes.append(n)
        else:
            raise ValueError('Invalid node type')

    def remove_node(self, node: Node):
        self.nodes.remove(node)
        if node in self.edges:
            for neighbor in self.edges[node]:
                self.edges[neighbor].remove(node)
            del self.edges[node]

    @property
    def is_crystal(self):
        self.crystal = True

    def add_edge(self, node1: Node, node2: Node):
        if node1 not in self.edges:
            self.edges[node1] = []
            self.edges[node1].append(node2)
        else:
            if node2 not in self.edges[node1]:
                self.edges[node1].append(node2)

        if node2 not in self.edges:
            self.edges[node2] = []
            self.edges[node2].append(node1)
        else:
            if node1 not in self.edges[node2]:
                self.edges[node2].append(node1)

    @property
    def graph_formula(self):
        symbols = list(map(lambda x: x.symbol, self.nodes))
        symb, count = np.unique(symbols, return_counts=True)
        return ''.join(map(lambda x: x[0] + str(x[1]), zip(symb, count)))

    @property
    def construct_graph(self) -> None:
        coordinates = np.array(list(map(lambda x: x.coordinate, self.nodes)))
        radii = np.array(list(map(lambda x: x.radius, self.nodes)))
        radii = np.tile(radii, (len(radii), 1))
        radii = self.scaler * (radii + radii.T)
        distances = cdist(coordinates, coordinates)
        distances -= radii
        # find the indexes of distances that are less than 0
        idx = np.where(distances <= 0)
        idx = list(zip(idx[0], idx[1]))
        idx = list(filter(lambda x: x[0] != x[1], idx))
        list(map(lambda x: self.add_edge(self.nodes[x[0]], self.nodes[x[1]]), idx))

    def _edge_vector_2d(self, node1: Node, node2: Node) -> np.ndarray:
        return np.array([node1.xy_project, node2.xy_project])

    def _edge_vector_3d(self, node1: Node, node2: Node) -> np.ndarray:
        return np.array([node1.coordinate, node2.coordinate])

    @property
    def plot(self):
        self.construct_graph
        nodes = np.array(list(map(lambda x: x.xy_project, self.nodes)))
        edges = list(map(lambda x: list(map(lambda y: self._edge_vector_2d(x, y), self.edges[x])), self.edges))
        edges = np.array([item for sublist in edges for item in sublist]).reshape(-1, 2, 2)
        idcs = np.array(list(map(lambda x: x.idx, self.nodes)))

        fig, ax = plt.subplots(figsize=(5, 5))
        for edge in edges:
            ax.plot(edge[:, 0], edge[:, 1], c='k', zorder=1)
        for node, idx in zip(nodes, idcs):
            ax.scatter(node[0], node[1], s=100, c='w', zorder=2, edgecolors='k')
            ax.text(node[0], node[1], str(idx), fontsize=6, ha='center', va='center', zorder=2)
        ax.set_aspect('equal')
        ax.set_xlabel('x (Å)')
        ax.set_ylabel('y (Å)')
        fig.tight_layout()
        fig.savefig('molecular_graph.png', dpi=600)

    @property
    def plot3d(self):
        self.construct_graph
        nodes = np.array(list(map(lambda x: x.coordinate, self.nodes)))
        edges = list(map(lambda x: list(map(lambda y: self._edge_vector_3d(x, y), self.edges[x])), self.edges))
        edges = np.array([item for sublist in edges for item in sublist]).reshape(-1, 2, 3)

        # 3d plot
        fig, ax = plt.subplots(subplot_kw={'projection': '3d'}, figsize=(5, 5))
        for node in nodes:
            ax.scatter(node[0], node[1], node[2], s=30, c='w', edgecolors='k', zorder=2)
        for edge in edges:
            ax.plot(edge[:, 0], edge[:, 1], edge[:, 2], c='k', zorder=1)
        ax.set_xlabel('x (Å)')
        ax.set_ylabel('y (Å)')
        ax.set_zlabel('z (Å)')
        plt.show()

    @property
    def get_centered_node(self) -> Node:
        for node in self.nodes:
            if np.allclose(node.coordinate, np.zeros(3)):
                return node
        raise ValueError('No centered node found')

    def extract_fragment(self, node: Node):
        seen = {node}
        mol_graph = MolGraph()
        sys.setrecursionlimit(1000000)

        if len(self.edges) == 0:
            raise ValueError('Graph is not constructed')

        def dfs(node):
            mol_graph.add_node(node)
            if node in self.edges:
                for neighbor in self.edges[node]:
                    if neighbor not in seen:
                        seen.add(neighbor)
                        dfs(neighbor)

        dfs(node)
        mol_graph.construct_graph
        return mol_graph

    def coordination_sphere(self, node: Node, cs_number: int):
        seen = {node}
        sphere_region = MolGraph()

        if len(self.edges) == 0:
            raise ValueError('Graph is not constructed')

        def add_coordination(node, rucursive_depth=0):
            sphere_region.add_node(node)

            if rucursive_depth == cs_number:
                return

            if node in self.edges:
                for neighbor in self.edges[node]:
                    if neighbor not in seen:
                        seen.add(neighbor)
                        add_coordination(neighbor, rucursive_depth + 1)

        add_coordination(node)
        sphere_region.construct_graph
        return sphere_region

    def bridge_nodes(self, node1: Node, node2: Node):
        seen = {node1, node2}
        fragment = []
        atom_dist = node1 - node2
        print("distance between node1 and node2: ", atom_dist)
        symbol = node1.symbol

        fragment.append(node1)
        fragment.append(node2)

        fcs1 = self.edges[node1]
        fcs2 = self.edges[node2]

        fcs1 = list(filter(lambda x: x.symbol != symbol, fcs1))
        fcs2 = list(filter(lambda x: x.symbol != symbol, fcs2))

        fragment.extend(fcs1)
        fragment.extend(fcs2)
        seen.update(fcs1)
        seen.update(fcs2)

        mol_graph = MolGraph()
        for node in fragment:
            mol_graph.add_node(node)
        mol_graph.construct_graph
        return mol_graph

    def _shift_node(self, node: Node):
        x, y, z = node.coordinate
        x_lim, y_lim, z_lim = self.boundaries / 2
        tol = 1.5
        if np.allclose(x, -x_lim, atol=tol) or np.allclose(x, x_lim, atol=tol):
            x = -x
        if np.allclose(y, -y_lim, atol=tol) or np.allclose(y, y_lim, atol=tol):
            y = -y
        if np.allclose(z, -z_lim, atol=tol) or np.allclose(z, z_lim, atol=tol):
            z = -z

        new_node = Node(np.array([x, y, z]), node.symbol, node.radius, node.idx, node.mass)
        self.add_node(new_node)
        self.remove_node(node)

    @property
    def fragment_graph(self):
        fragments = []
        seen = set()
        for node in self.nodes:
            if node.idx not in seen:
                graph = self.extract_fragment(node)
                nodes = graph.nodes
                idx = set(map(lambda x: x.idx, nodes))
                seen.update(idx)
                fragments.append(graph)
            else:
                continue
        return fragments

    @staticmethod
    def mass_center(coords: np.ndarray, masses: np.ndarray) -> np.ndarray:
        center = np.sum(coords * masses[:, None], axis=0) / np.sum(masses)
        coords -= center
        tensor = coords.T @ np.diag(masses) @ coords
        eigval, eigvec = np.linalg.eigh(tensor)
        idx = eigval.argsort()[::-1]
        idx = idx[::-1]
        eigval = eigval[idx]
        prin_axes = eigvec[:, idx]
        coords = coords @ prin_axes
        return coords

    def to_xyz(self, filename: str=None, mass_centered: bool=False):
        if filename is None:
            filename = self.graph_formula + '.xyz'
        coordinates = np.array(list(map(lambda x: x.coordinate, self.nodes)))
        symbols = list(map(lambda x: x.symbol, self.nodes))
        if mass_centered:
            masses = np.array(list(map(lambda x: x.mass, self.nodes)))
            coordinates = self.mass_center(coordinates, masses)
        xyz_py.save_xyz(filename, symbols, coordinates)

    def graph_compare(self, other: 'MolGraph') -> bool:
        if len(self.nodes) != len(other.nodes):
            return False

        symbols1 = list(map(lambda x: x.symbol, self.nodes))
        symbols2 = list(map(lambda x: x.symbol, other.nodes))
        s1, c1 = np.unique(symbols1, return_counts=True)
        s1dict = dict(zip(s1, c1))
        s2, c2 = np.unique(symbols2, return_counts=True)
        s2dict = dict(zip(s2, c2))

        if s1dict != s2dict:
            return False

        mass1 = np.array(list(map(lambda x: x.mass, self.nodes)))
        coords1 = np.array(list(map(lambda x: x.coordinate, self.nodes)))
        coords1 = self.mass_center(coords1, mass1)

        mass2 = np.array(list(map(lambda x: x.mass, other.nodes)))
        coords2 = np.array(list(map(lambda x: x.coordinate, other.nodes)))
        coords2 = self.mass_center(coords2, mass2)

        rmsd = xyz_py.minimise_rmsd(coords1, coords2)[0]
        if rmsd < 1:
            return True
        else:
            return False



import numpy as np

from .parser import ParseOUTCAR

THZ_TO_INV_CM = 33.35641


def parse_frequencies(outcar: ParseOUTCAR):
    freqs = np.array(list(outcar.hessian_eigenvalues))
    if len(freqs) == 0:
        raise ValueError("No eigenvalues found in OUTCAR \n \
                Run VASP with the following tags: \n \
                IBRION=5, NSW=1 \n\n \
                NSW=1 is important to calculate the normal modes in older versions of VASP \n \
                It is suggested to increase the convergence tolerance EDIFFG to 1E-6")
    return freqs * THZ_TO_INV_CM


def parse_displacements(outcar: ParseOUTCAR):
    displacements = np.array(list(outcar.hessian_eigenvectors))
    disp_norms = list(map(np.linalg.norm, displacements))
    if not all(np.isclose(1., disp_norms)):
        raise ValueError("Displacements are not normalized")
    return displacements



"""
A module for parsing vasp output files.
"""

import os
import re

from .structure import Structure


class Parser(Structure):
    """
    Parent class for all parsers
    """

    def __init__(self, output_file, poscar_file='POSCAR'):
        """
        Initializes the parser object.
        """
        # Initialize the structure object
        if not os.path.exists(output_file):
            raise ValueError("File not found: {}".format(output_file))
        if not (os.path.exists(poscar_file)):
            raise IOError("File not found: {}".format(poscar_file))
        poscar = Structure.from_poscar(poscar_file)
        for atrr in poscar.__dict__:
            setattr(self, atrr, getattr(poscar, atrr))
        self.output_file = output_file
        self.poscar_file = poscar_file

    @property
    def read_lines(self):
        """
        Reads the lines from the output file
        """
        with open(self.output_file, 'r') as f:
            return f.readlines()


class ParseOUTCAR(Parser):
    """
    Parses the OUTCAR file
    """

    # Patterns

    external_pressureREGEX = re.compile(
            r'^\s+external\s+pressure\s+=\s+(?P<pressure>[+-]?\d+\.\d+)\s+'
            )

    elapsed_timeREGEX = re.compile(
            r'^\s+Elapsed\s+time\s+\(sec\)\:\s+(?P<time>\d+\.\d+)\s+'
            )

    eigenval_hessianREGEX = re.compile(
            r'^\s+\d+\s*f\s*=\s*(?P<eigenval>\d+\.\d+)\s+THz\s+'
            )

    eigenval_hessianREGEX2 = re.compile(
            r'^\s+\d+\s*f/i\s*=\s*(?P<eigenval>\d+\.\d+)\s+THz\s+'
            )

    def __init__(self, output_file='OUTCAR', poscar_file='POSCAR'):
        """
        Initializes the OUTCAR parser object.
        """
        super().__init__(output_file, poscar_file)

    @property
    def external_pressure(self):
        """
        Returns the external pressure
        """
        # UNITS: kB
        for line in self.read_lines:
            match = self.external_pressureREGEX.match(line)

            if match:
                yield float(match.group('pressure'))

    @property
    def elapsed_time(self):
        for line in self.read_lines:
            match = self.elapsed_timeREGEX.match(line)

            if match:
                yield float(match.group('time'))

    @property
    def hessian_eigenvalues(self):
        """
        Returns the hessian eigenvalues
        """
        for line in self.read_lines:
            match = self.eigenval_hessianREGEX.match(line)
            match2 = self.eigenval_hessianREGEX2.match(line)

            if match:
                yield float(match.group('eigenval'))

            if match2:
                yield float(match2.group('eigenval'))

    @property
    def hessian_eigenvectors(self):
        """
        Returns the hessian eigenvectors
        """
        for idx, line in enumerate(self.read_lines):
            match = self.eigenval_hessianREGEX.match(line)
            match2 = self.eigenval_hessianREGEX2.match(line)

            if match:
                eigvecs = self.read_lines[idx+2:idx+2+self.N]
                eigvecs = list(map(lambda x: x.strip().split(), eigvecs))
                eigvecs = list(map(lambda x: list(map(lambda y: float(y), x[3:])), eigvecs))
                yield eigvecs

            if match2:
                eigvecs = self.read_lines[idx+2:idx+2+self.N]
                eigvecs = list(map(lambda x: x.strip().split(), eigvecs))
                eigvecs = list(map(lambda x: list(map(lambda y: float(y), x[3:])), eigvecs))
                yield eigvecs


class ParseOSZICAR(Parser):
    """
    Parses the OSZICAR file
    """

    energyRegex = re.compile(
            r"^\s+\d+\s+F=\s+[+-]?(\d*)?\.\d*[eE]?[+-]?\d*\s+E0=\s+(?P<e_0>[+-]?(\d*)?\.\d*[eE]?[+-]?\d*)\s+"
            )

    stepREGEX = re.compile(
            r'[A-Z]{3}:\s+(?P<step>\d+)\s+'
            )

    def __init__(self, ouput_file='OSZICAR', poscar_file='POSCAR'):
        """
        Initializes the OSZICAR parser object.
        """
        super().__init__(ouput_file, poscar_file)

    @property
    def energy(self):
        """
        returns the energy per atom
        """
        for line in self.read_lines:
            match = self.energyRegex.match(line)

            if match:
                yield float(match.group('e_0'))

    @property
    def electronic_steps(self):
        """
        returns the number of electronic steps per
        ionic step
        """
        steps = []
        for line in self.read_lines:
            match = self.stepREGEX.match(line)

            if match:
                steps.append(int(match.group('step')))
        electronic_steps = []
        prev = 0
        for idx, step in enumerate(steps):
            if step == 1:
                electronic_steps.append(steps[prev:idx])
                prev = idx
        electronic_steps.append(steps[prev:])
        for list in electronic_steps:
            if len(list) < 1:
                electronic_steps.remove(list)
        return [max(x) for x in electronic_steps]



# Import argparse
import argparse
from textwrap import dedent
import numpy as np
import shutil
import os
import pandas as pd

# Import Modules
from . import core
from . import phonon
from . import conv
from . import input
from . structure import Molecule
# Create the wrapper funstions

THZ_TO_INV_CM = 33.35641
C0 = 299792458


def generate_input_func(args):
    '''
    Wrapper function for the generate_input function

    parameters
    ----------
    filename : str
    config_file : file

    Returns
    -------
    INCAR : file
    '''
    core.generate_input(
        filename=args.poscar,
        calculation=args.config_file,
            )

    with open('INCAR', 'a') as f:
        f.write('ENCUT = {}\n'.format(args.encut))

    if args.mesh is not None:
        core.generate_kpoints(mesh=args.mesh)
    elif args.kspacing is not None:
        with open('INCAR', 'a') as f:
            f.write('KSPACING = {}\n'.format(args.kspacing))
    else:
        raise NotImplementedError('No kpoint scheme specified')


def generate_supercell_func(args):
    '''
    Wrapper function for the generate_supercell function

    parameters
    ----------
    expansion : list
    poscar : str

    Returns
    -------
    POSCAR : file
    '''
    core.generate_supercell(
            expansion=np.array(args.dim, dtype=int),
            filename=args.poscar,
            )


def generate_job_func(args):
    '''
    Wrapper function for the generate_job function

    parameters
    ----------
    title : str
    cores : int
    vasp_type : str

    Returns
    -------
    job.sh : file
    '''
    core.generate_job(
            title=args.title,
            cores=args.cores,
            vasp_type=args.vasp_type,
            job_title=args.output,
            array=None,
            array_file=None,
            )


def start_up_func(args):
    '''
    Wrapper function for the start_up function

    parameters
    ----------
    None

    Returns
    -------
    configuration files : file
    '''
    core.create_input_configurations()


def dope_structure_func(args):
    '''
    Wrapper function for the dope_structure function

    parameters
    ----------
    filename : str
    dopant : str
    replace : str
    instances : int

    Returns
    -------
    POSCARs : file
    '''
    core.dope_structure_sym(
        filename=args.poscar,
        dopant=args.dopant,
        replace=args.replace,
        instances=args.instances,
        no_sym=args.no_symm,
        )


def generate_defect_func(args): 
    '''
    Wrapper function for the generate_defect function

    parameters
    ----------
    filename : str
    site : str
    instances : int

    Returns
    -------
    POSCAR : file
    '''
    core.generate_defect(
            filename=args.poscar,
            site=args.site,
            instances=args.instances,
            )


def find_molecule_func(args):
    '''
    Wrapper function for the asymmetric_unit function

    parameters
    ----------
    filename : str
    atom : str
    bond_max : float

    Returns
    -------
    POSCAR : file
    '''
    core.molecule(
            filename=args.poscar,
            atom=args.atom,
            scaler=args.scaler,
            )


def convert_cif_func(args):
    '''
    Wrapper function for the convert_cif function

    parameters
    ----------
    filename : str

    Returns
    -------
    POSCAR : file
    '''
    core.convert_cif(
            filename=args.filename,
            )

def phonon_calc(args):
    '''
    wrapper function for phonon calculation

    parameters
    ----------
    supercell : list
    mesh : list
    encut : int
    cores : int
    bec : bool

    Returns
    -------
    None
    '''
# Check if the config file exists
    if not phonon.check_config():
        raise ValueError('No config file found, please run the start_up command')

# Generate the input files
    core.generate_input('POSCAR', 'phonon')
    core.generate_kpoints(mesh=args.k_point_mesh)

    with open('INCAR', 'a') as f:
        f.write('ENCUT = {}\n'.format(str(args.encut)))

# write the phonon files to a phonopy_file
    phonon.store_disp_files()

# Generate the job file
    core.generate_job(
            title='phonon',
            cores=args.cores,
            vasp_type=args.vasp_type,
            job_title='phonopy',
            array='phonopy',
            array_file='phonon_files.txt',
    )

    if args.bec:
        os.mkdir('BEC')
        os.chdir('BEC')
        shutil.copy('../POSCAR', '.')
        core.generate_input('POSCAR', 'bec')
        core.generate_kpoints(mesh=args.bec_mesh)

        with open('INCAR', 'a') as f:
            f.write('ENCUT = {}\n'.format(str(args.encut)))
        core.generate_job('BEC', args.bec_cores, 'standard',
                          job_title='BORN', array=None, array_file=None)
        os.chdir('..')


def convergence_func(args):
    """
    Wrapper function for the convergence function

    parameters
    ----------
    poscar : str
        The structure file to be used, POSCAR or CIF
    cores : int
        The number of cores to be used
    """
    if not os.path.exists(args.poscar):
        raise FileNotFoundError('{} does not exist'.format(args.poscar))

    if args.gen:
        input_file = input.InputFileGenerator(filename=args.poscar, calculation=args.scf_type)
        input_file.generate_incar()
        input_file.generate_potcar()
        combinations = conv.get_combinations(filename=args.poscar, max_encut=args.max_encut + 50, min_encut=args.min_encut)
        conv.make_dirs(combinations)
        core.generate_job(title='CONV', cores=args.cores, vasp_type='standard', job_title='CONV', array='convergence', array_file='convergence.txt')
    elif args.parse:
        combinations = conv.get_combinations(filename=args.poscar, max_encut=args.max_encut + 50, min_encut=args.min_encut)
        conv.extract_data(combinations)
    elif args.plot:
        if not os.path.exists('conv.csv'):
            raise FileNotFoundError('conv.csv does not exist, please run the parse command first')
        data = pd.read_csv('conv.csv')
        if not os.path.exists(args.poscar):
            raise FileNotFoundError('{} does not exist'.format(args.poscar))
        struct = core.Structure.from_poscar(args.poscar)
        num_atoms = struct.N
        conv.kpoint_plot(data, num_atoms, args.gradient)
        conv.cutoff_plot(data, num_atoms, args.gradient)
    else:
        raise NotImplementedError('No action specified, please use the --gen, --parse or --plot flags')


def bridge_dimer_func(args):
    if not os.path.exists(args.poscar):
        raise FileNotFoundError('{} does not exist'.format(args.poscar))

    mol = Molecule(args.poscar, args.atom, args.scaler)
    mol.bridge_dimer()
    qm_graph = mol.qm_graph
    qm_graph.to_xyz("{}.xyz".format("".join(qm_graph.graph_formula)), mass_centered=True)


def coord_sphere_func(args):
    """
    Wrapper function for the coord_sphere function

    parameters
    ----------
    poscar : str
        The structure file to be used, POSCAR
    atom : str
        The atom to be used
    bond_max : float
        The maximum bond length
    coordination_number : int
        The number of coordination spheres to be found: (1, 2)
    """
    if not os.path.exists(args.poscar):
        raise FileNotFoundError('{} does not exist'.format(args.poscar))

    core.coord_sphere(
            filename=args.poscar,
            atom=args.atom,
            sphere_num=args.coordination_sphere_number,
            )

def reduce_func(args):
    """
    Wrapper function for the reduce function

    parameters
    ----------
    poscar : str
        The structure file to be used, POSCAR

    primitive : bool
        Whether to reduce to primitive cell or not

    niggli: bool
        Whether to reduce to niggli cell or not

    delaunay: bool
        Whether to reduce to delaunay cell or not

    refine: bool
        Whether to refine the cell or not
    """

    if not os.path.exists(args.poscar):
        raise FileNotFoundError('{} does not exist'.format(args.poscar))

    core.reduce(
            filename=args.poscar,
            primative=args.primitive,
            niggli=args.niggli,
            delaunay=args.delaunay,
            refine=args.refine,
            )

# phonopy interface
def phonopy_func(args):
    """
    function for interfacing with phonopy
    includes:
        - band structure input file generation
        - dos input file generation
    """
    if not os.path.exists(args.poscar):
        raise FileNotFoundError('{} does not exist'.format(args.poscar))

    if args.band and not args.mesh:
        if args.dim is None:
            raise ValueError('Please specify the dimensions of the supercell (--dim)')
        core.write_band(args.poscar, args.dim)

    if args.mesh and not args.band:
        if args.dim is None:
            raise ValueError('Please specify the dimensions of the supercell (--dim)')
        if args.mesh_points is None:
            raise ValueError('Please specify the mesh points (--mesh_points)')
        core.write_mesh(args.poscar, args.dim, args.mesh_points)

    if args.band and args.mesh:
        if args.dim is None:
            raise ValueError('Please specify the dimensions of the supercell (--dim)')
        if args.mesh_points is None:
            raise ValueError('Please specify the mesh points (--mesh_points)')
        core.write_band_dos(args.poscar, args.dim, args.mesh_points)


###########


def read_args(arg_list=None):
    '''Reads the command line arguments'''
    parser = argparse.ArgumentParser(
            prog='vasp_suite',
            description=dedent('''A suite of tools for VASP calculations'''),
            epilog=dedent('''To display options for a specific programme, use vasp_suite <programme> -h'''),
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    # Subparsers
    subparsers = parser.add_subparsers(dest='prog')

    gen_inp = subparsers.add_parser(
            'generate_input',
            description=dedent(
                '''
                Generation of INCAR and POTCAR files for VASP calculations.
                '''
                ),
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    gen_inp.set_defaults(func=generate_input_func)

    gen_inp.add_argument(
            '--config_file',
            required=True,
            help=dedent(
                """
                The configuration file for the input generation.

                Example:

                Inside ~/.vasp_suite_templates '.ini' configuration files are
                stored. To perform a relaxation caluclation using the relaxation.ini
                template, use the following command:

                "vasp_suite generate_input relaxation"
                """
                )
            )

    gen_inp.add_argument(
            '--poscar',
            required=True,
            help=dedent(
                '''
                The name of the structure file.
                '''
                ),
            )

    gen_inp.add_argument(
            '--encut',
            required=True,
            type=int,
            help=('The ENCUT value to be used in the calculation')
            )

    kp = gen_inp.add_mutually_exclusive_group(required=True)

    kp.add_argument(
            '--mesh',
            nargs=3,
            metavar=('Nx', 'Ny', 'Nz'),
            type=int,
            help='The kpoint mesh to be used in the calculation'
    )

    kp.add_argument(
        '--kspacing',
        type=float,
        help='The kpoint spacing to be used in the INCAR'
    )

    gen_job = subparsers.add_parser(
        'generate_job',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_job.set_defaults(func=generate_job_func)

    gen_job.add_argument(
        '--title',
        help='The title of the job'
        )

    gen_job.add_argument(
        '--cores',
        help='The number of cores to use',
        type=int,
        )

    gen_job.add_argument(
        '--vasp_type',
        help='The vasp program to use',
        choices=['standard', 'gamma'],
        required=True,
        )

    gen_job.add_argument(
            '--output',
            help='The output file name',
            required=False,
            default='submit'
            )

    gen_supercell = subparsers.add_parser(
        'supercell',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_supercell.set_defaults(func=generate_supercell_func)

    gen_supercell.add_argument(
        '--dim',
        nargs=3,
        metavar=('Nx', 'Ny', 'Nz'),
        help=dedent(
            '''
            The expansion vector for the supercell, a b c
            '''
            ),
        )

    gen_supercell.add_argument(
            '--poscar',
            help=dedent(
                '''
                The name of the structure file
                '''
                ),
            )

    start_up = subparsers.add_parser(
        'set_up',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    start_up.set_defaults(func=start_up_func)

    dope_struct = subparsers.add_parser(
        'dope_structure',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    dope_struct.set_defaults(func=dope_structure_func)

    dope_struct.add_argument(
        '--poscar',
        help='The name of the structure file',
        )

    dope_struct.add_argument(
            '--dopant',
            help='The element to dope the structure with',
            )

    dope_struct.add_argument(
            '--replace',
            help='The element to replace',
            )

    dope_struct.add_argument(
            '--instances',
            help='The number of instances of the dopant to add',
            type=int,
            )

    dope_struct.add_argument(
            '--no_symm',
            help='Do not remove symmetry equivelent structures',
            action='store_true',
            )

    gen_defect = subparsers.add_parser(
        'generate_defect',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_defect.set_defaults(func=generate_defect_func)

    gen_defect.add_argument(
            '--poscar',
            help='The name of the structure file',
            )

    gen_defect.add_argument(
            '--site',
            help='The name of the atom to remove',
            )

    gen_defect.add_argument(
            '--instances',
            help='The number of instances of defect',
            type=int,
            )

    asym = subparsers.add_parser(
        'find_molecule',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    asym.set_defaults(func=find_molecule_func)

    asym.add_argument(
            '--poscar',
            help='The name of the structure file',
            required=True,
            )

    asym.add_argument(
            '--atom',
            help='The name of the spin centre in the molecular crystal',
            required=True,
            )

    asym.add_argument(
            '--scaler',
            help='The scaling factor when finding connectivity',
            type=float,
            default=1.12,
            )

    cif = subparsers.add_parser(
            'convert_cif',
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    cif.set_defaults(func=convert_cif_func)

    cif.add_argument(
            '--filename',
            help='The name of the .CIF file',
            required=True,
            )

    phonon = subparsers.add_parser(
            'phonopy_job',
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    phonon.set_defaults(func=phonon_calc)

    std = phonon.add_argument_group('Phonopy args')

    std.add_argument(
            '--k_point_mesh',
            help='The mesh to use for the phonon calculation',
            nargs=3,
            metavar=('Nx', 'Ny', 'Nz'),
            type=int,
            )

    std.add_argument(
            '--cores',
            help='The number of cores to use for the phonon calculation',
            type=int,
            required=True,
            )

    std.add_argument(
            '--encut',
            help='The energy cutoff for the phonon calculation',
            type=int,
            )

    std.add_argument(
            '--vasp_type',
            help='The vasp program to use',
            choices=['standard', 'gamma'],
            required=True,
            )

    bec = phonon.add_argument_group('Born effective charges')

    bec.add_argument(
            '--bec',
            help='Calculate the Born effective charges',
            action='store_true',
            )

    bec.add_argument(
            '--bec_mesh',
            help='The mesh to use for the Born effective charges',
            nargs=3,
            metavar=('Nx', 'Ny', 'Nz'),
            type=int,
            default=[1, 1, 1],
            )

    bec.add_argument(
            '--bec_cores',
            help='The number of cores to use for the BEC calculation',
            type=int,
            )

    conv = subparsers.add_parser(
            "convergence",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )

    conv.set_defaults(func=convergence_func)

    prog_conv = conv.add_argument_group('Convergence programs')
    conv_prog = prog_conv.add_mutually_exclusive_group(required=True)

    conv_prog.add_argument(
            '--gen',
            help='Generate the convergence testing directories, input files and submission script',
            action='store_true',
            )

    conv_prog.add_argument(
            '--parse',
            help='Parse the convergence testing directories',
            action='store_true',
            )

    conv_prog.add_argument(
            '--plot',
            help='Plot the convergence testing results',
            action='store_true',
            )

    plot = conv.add_argument_group('Plotting options')
    plot.add_argument(
            '--gradient',
            help='Plot the gradient of the convergence testing results',
            action='store_true',
            )

    conv_std = conv.add_argument_group('Standard args')
    conv_std.add_argument(
            '--poscar',
            help='The name of the POSCAR file',
            required=True,
            )

    conv_std.add_argument(
            '--cores',
            help='The number of cores to use for the calculation',
            type=int,
            default=1,
            )

    conv_std.add_argument(
            '--max_encut',
            help='The maximum energy cutoff to test',
            type=int,
            default=1000,
            )

    conv_std.add_argument(
            '--min_encut',
            help='The minimum energy cutoff to test',
            type=int,
            default=200,
            )

    conv_std.add_argument(
            '--scf_type',
            help='The config file for the scf calculation. Default is scf with VDW',
            type=str,
            default='scf',
            required=False,
            )

    dimer = subparsers.add_parser(
            "bridge_dimer",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )

    dimer.set_defaults(func=bridge_dimer_func)

    dimer.add_argument(
            "--poscar",
            help="The name of the POSCAR file",
            required=True,
            )

    dimer.add_argument(
            "--atom",
            help="The name of the first atom in the dimer e.g. Dy1",
            required=True,
            )

    dimer.add_argument(
            "--scaler",
            help="The scaling factor for the dimer",
            type=float,
            default=1.1,
    )

    cd_sphere = subparsers.add_parser(
            "coordination_sphere",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )

    cd_sphere.set_defaults(func=coord_sphere_func)

    cd_sphere.add_argument(
            '--poscar',
            help='The name of the POSCAR file',
            )

    cd_sphere.add_argument(
            '--atom',
            help='The name of the atom to find the coordination sphere of, e.g. "Dy1"',
            )

    cd_sphere.add_argument(
            '--coordination_sphere_number',
            help='The coordination sphere number to find',
            type=int,
            )

    phon_int = subparsers.add_parser(
            "phonopy",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )

    phon_int.set_defaults(func=phonopy_func)

    plot_args = phon_int.add_argument_group('Plotting options')
    plot_args.add_argument(
            '--band',
            help='Write the band.conf file?',
            action='store_true',
            )

    plot_args.add_argument(
            '--mesh',
            help='Write the mesh.conf file?',
            action='store_true',
            )

    phon_std = phon_int.add_argument_group('Standard args')
    phon_std.add_argument(
            '--poscar',
            help='The name of the POSCAR file',
            )

    phon_std.add_argument(
            '--dim',
            help='The supercell dimensions',
            metavar=('Nx', 'Ny', 'Nz'),
            nargs=3,
            type=int,
            )

    phon_std.add_argument(
            '--mesh_points',
            help='The mesh points',
            metavar=('Nx', 'Ny', 'Nz'),
            nargs=3,
            type=int,
            )

    red = subparsers.add_parser(
            "reduce",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )

    red.set_defaults(func=reduce_func)

    red.add_argument(
            '--poscar',
            help='The name of the POSCAR file',
            )

    red_type = red.add_mutually_exclusive_group(required=True)

    red_type.add_argument(
            '--primitive',
            help='Reduce to the primitive cell',
            action='store_true',
            )

    red_type.add_argument(
            '--niggli',
            help='Reduce to the Niggli cell',
            action='store_true',
            )

    red_type.add_argument(
            '--delaunay',
            help='Reduce to the Delaunay cell',
            action='store_true',
            )

    red_type.add_argument(
            '--refine',
            help='Refine the cell',
            action='store_true',
            )

    # Parse the ArgumentParser
    parser.set_defaults(func=lambda args: parser.print_help())
    args = parser.parse_known_args(arg_list)

    # Select programme
    if args in ['generate_input', 'generate_job', 'supercell', 'set_up',
                'dope_structure', 'generate_defect', 'convert_cif',
                'phonopy_job', 'convergence', 'coordination_sphere',
                'phonopy', 'reduce', 'find_molecule']:
        args.func(args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)


def main():
    read_args()



'''
A programme which holds the core functions of vasp_suite.
'''

# Imports
import numpy as np
import os

from .structure import Structure, Dope, Molecule
from .input import InputFileGenerator
from .submission import SubmissionWriter
# Functions


def generate_input(filename: str, calculation: str):
    '''
    Generates Input files for VASP Calculations.

    Parameters
    ----------
    filename : str
        The name of the input file to be generated.
    calculation : str
        The type of calculation to be performed.
    '''
    input_files = InputFileGenerator(filename, calculation)
    input_files.generate_incar()
    input_files.generate_potcar()


def generate_job(title, cores, vasp_type, job_title, array, array_file):
    job = SubmissionWriter(title, cores, vasp_type)
    hostname = job.hostname
    with open('{}.sh'.format(job_title), 'w') as Submit:
        if hostname == 'csf3':
            job.write_csf3_header(Submit)
            if array is None:
                job.write_module_load(Submit)
                job.write_body(Submit)
            else:
                job.write_array_header_csf3(array_file, Submit)
                if array.lower() == 'phonopy':
                    job.write_phonopy_body(Submit)
                elif array.lower() == 'convergence':
                    job.write_convergence_body(Submit)
                else:
                    raise NotImplementedError('Array job not implemented')
        elif hostname == 'csf4':
            job.write_csf4_header(Submit)
            if array is None:
                job.write_module_load(Submit)
                job.write_body(Submit)
            else:
                job.write_array_header_csf4(array_file, Submit)
                if array.lower() == 'phonopy':
                    job.write_phonopy_body(Submit)
                elif array.lower() == 'convergence':
                    job.write_convergence_body(Submit)
                else:
                    raise NotImplementedError('Array job not implemented')

    os.system('chmod +x {}.sh'.format(job_title))
    if hostname == 'csf3':
        print('Submit job with: qsub {}.sh'.format(job_title))
    if hostname == 'csf4':
        print('Submit job with: sbatch {}.sh'.format(job_title))


def generate_supercell(expansion: np.ndarray, filename: str):
    '''
    Generates a supercell from a given expansion matrix.

    Parameters
    ----------
    expansion : np.ndarray
        The expansion of each lattice vector.
    filename : str
        The name of the file to be expanded.
    '''
    struct = Structure.from_poscar(filename)
    coords, lattice_vectors, atom_list = struct.generate_supercell(n=expansion)
    coords, atom_list, atoms, natoms = struct.reorder_supercell(coords, atom_list)
    struct.coords = coords
    struct.lattice_vectors = lattice_vectors
    struct.atom_list = atom_list
    struct.atoms = atoms
    struct.natoms = natoms
    struct.write_poscar('POSCAR_supercell')


def dope_structure_sym(filename: str, dopant: str, replace: str, instances: int, no_sym: bool):
    struct = Dope(filename, dopant, replace, instances)
    struct.generate_structures()
    if not no_sym:
        struct.symmetrize()
    struct.write_poscars()


def generate_defect(filename: str, site: str, instances: int):
    '''
    Generates a defect structure from a given structure.

    Parameters
    ----------
    filename : str
        The name of the file.
    site : str
        The site to be removed.
    instances : int
        The number of instances of the defect.
    '''
    _d = Dope(filename, 'D', site, instances)
    _d.generate_structure()
    _d.Create_defect()
    _d.write_poscars()


def calculate_kpoints(filename: str) -> None:
    '''
    Calculates possible kpoint meshes for a given structure.
    
    Parameters
    ----------
    filename : str
        The name of the file to be analysed.
    '''
    struct = Structure.from_poscar(filename)
    mesh, density = struct.calculate_mesh()
    for i in range(len(mesh)):
        print(f'Mesh: {mesh[i][0]} {mesh[i][1]} {mesh[i][2]} Density: {density[i]}')
    return None

def generate_kpoints(mesh: list) -> None:
    '''
    Generates a KPOINTS file for a given mesh.

    Parameters
    ----------
    mesh : list
        The mesh to be used.
    '''
    with open('KPOINTS', 'w') as f:
        f.write('Regular {}\t{}\t{} gamma centred mesh\n'.format(*mesh))
        f.write('0\nGamma\n')
        f.write('{}\t{}\t{}\n'.format(*mesh))
        f.write('0 0 0')

def convert_cif(filename: str) -> None:
    '''
    Converts a cif file to a POSCAR file.

    Parameters
    ----------
    filename : str
        The name of the file to be converted.
    '''
    struct = Structure.from_cif(filename)
    struct.write_poscar('POSCAR')


def molecule(filename: str, atom: str, scaler: float):
    '''
    Finds a molecule in a structure.

    Parameters
    ----------
    filename : str
        The name of the file to be analysed.
    atom : str
        The atom to build the molecule from.
    bond_max : float
        The maximum bond length.
    '''
    struct = Molecule(filename, atom, scaler)
    struct.find_molecule()


def coord_sphere(filename: str, atom: str, sphere_num: int):
    '''
    Finds the coordination sphere of a given atom.

    Parameters
    ----------
    filename : str
        The name of structure file
    atom : str
        The central atom
    bond_max : float
        The maximum bond length
    sphere_num : int
        The coordination sphere to be found
    '''
    struct = Molecule(filename, atom, scaler=1.1)
    struct.get_coordination_sphere(sphere_num)
    struct.qm_graph.to_xyz("{}.xyz".format("".join(struct.qm_graph.graph_formula), mass_center=True))


def write_band(filename: str, dim: list):
    poscar = Structure.from_poscar(filename)
    kpath = poscar.get_kpath()
    path = [kpoint[1] for kpoint in kpath]
    path = [' '.join([str(i) for i in kpoint]) for kpoint in path]
    band_labels = [kpoint[0] for kpoint in kpath]
    band_labels = list(map(lambda x: x.replace('GAMMA', '$\Gamma$'), band_labels))

    with open('band.conf', 'w') as f:
        f.write('ATOM_NAME = {}\n'.format(' '.join(poscar.atoms)))
        f.write('DIM = {} {} {}\n'.format(*dim))
        f.write('BAND = {}\n'.format('  '.join(path)))
        f.write('BAND_LABELS = {}\n'.format(' '.join(band_labels)))

    print('\nK-path written to band.conf')
    print('\tSpacegroup: {}'.format(poscar.spacegroup))
    print('\tAtoms: {}\n'.format(' '.join(poscar.atoms)))
    print('\tk-path:')
    print('\t{}'.format(' -> '.join(band_labels)))
    for kpoint in kpath:
        print('\t{}'.format(kpoint[1]))


def write_mesh(filename: str, dim: list, mesh_points: list):
    poscar = Structure.from_poscar(filename)
    atoms = ' '.join(poscar.atoms)

    with open('mesh.conf', 'w') as f:
        f.write('ATOM_NAME = {}\n'.format(atoms))
        f.write('DIM = {} {} {}\n'.format(*dim))
        f.write('MESH = {} {} {}\n'.format(*mesh_points))


def write_band_dos(filename: str, dim: list, mesh_points: list):
    write_band(filename, dim)
    write_mesh(filename, dim, mesh_points)

    with open('band-pdos.conf', 'w') as f:
        with open('band.conf', 'r') as band:
            f.write(band.read())
        with open('mesh.conf', 'r') as mesh:
            f.write(mesh.readlines()[2])


def reduce(filename, primative, niggli, delaunay, refine):
    struct = Structure.from_poscar(filename)
    cell = struct.reduced_cell(primative, niggli, delaunay, refine)
    struct.coords = cell[1]
    struct.lattice_vectors = cell[0]
    struct.natoms = [list(cell[2]).count(atom) 
                     for atom in list(set(cell[2]))]
    struct.write_poscar('POSCAR_reduced')


def create_input_configurations() -> None:
    '''
    Creates input configurations for VASP calculations.
    '''
    if not os.path.exists(os.path.expanduser('~/.vasp_suite_configs')):
        os.mkdir(os.path.expanduser('~/.vasp_suite_configs'))

    cwd = os.getcwd()
    os.chdir(os.path.expanduser('~/.vasp_suite_configs'))
    for key, value in default_configs.items():
        with open('{}.ini'.format(key), 'w') as f:
            f.write(value)
    os.chdir(cwd)


default_configs = {
        'relaxation': """[General]
prec = ACCURATE
lreal = .FALSE.
lasph = .TRUE.
ismear = 0
sigma = 0.01
nelm = 100
nelmin = 4
ncore = 4
lwave = .FALSE.
lcharg = .FALSE.
lorbit = 11
ivdw = 11
[Convergence]
ediff = 1e-08
ediffg = -0.01
[Optimisation]
ibrion = 2
isif = 4
nsw = 100
potim = 0.5
[Functional]
gga = PE
""",
        'scf': """[General]
prec = ACCURATE
lreal = .FALSE.
lwave = .FALSE.
lasph = .TRUE.
lcharg = .FALSE.
ismear = 0
sigma = 0.01
nelm = 100
nelmin = 4
ncore = 4
ivdw = 11
[Convergence]
ediff = 1e-08
[SCF]
ibrion = -1
isif = 2
nsw = 0
[Functional]
gga = PE
""",
        'phonon': """[General]
prec = Accurate
algo = Normal
ediff = 1E-8
ismear = 0
sigma = 0.01
lasph = .TRUE.
lcharg = .FALSE.
lreal = .FLASE.
lwave = .FALSE.
nelm = 250
[SCF]
isif = 2
nsw = 0
[Functional]
gga = PE
""",
        'bec': """[BORN]
prec = Accurate
algo = Fast
sigma = 0.01
ediff = 1e-8
ismear = 0
ivdw = 11
lcharg = .FALSE.
lepsilon = .TRUE.
lreal = .FALSE.
lwave = .FALSE.
isif = 2
nelm = 250
nsw = 0
[Functional]
gga = PE
""",
}



'''
A module for the Structure class. Structure will perform operations structure data files.
'''
# Imports
import re
import os
import numpy as np
import scipy as sp
from functools import wraps
from time import perf_counter
from ase import io
import spglib
import seekpath

from ase.data import atomic_numbers, atomic_masses, covalent_radii

from .graphs import MolGraph, Node


def timer(func):
    '''
    A decorator function to time functions.
    '''
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = perf_counter()
        print(f'Running {func.__name__}...')
        func(*args, *kwargs)
        end = perf_counter()
        print(f"Time elapsed: {end - start}")
    return wrapper


class Structure:
    '''
    A class of functions for manipulating structure data files.
    '''

    def __init__(self, name, scale, av, _atom_list, _N, _type,
                 coords, atoms, natoms, cart_coords, bv, V):
        '''
        Initializes the Structure class.
        '''
        self.name = name
        self.scale = scale
        self.lattice_vectors = av
        self.atom_list = _atom_list
        self.N = _N
        self._type = _type
        self.coords = coords
        self.atoms = atoms
        self.natoms = natoms
        self.cart_coords = cart_coords
        self.recip_vectors = bv
        self.V = V

    @classmethod
    def from_poscar(cls, filename):
        '''
        Reads and formats a .vasp, POSCAR or CONTCAR file.

        Parameters
        ----------
        filename : str

        Returns
        -------
        Structure object
        '''
        with open(filename) as f:
            data = f.readlines()
        data = [x.strip().split() for x in data]

        name: str = data[0]
        scale: float = float(data[1][0])
        av: np.ndarray = np.array(data[2:5], dtype=float)
        atoms: list = data[5]
        natoms: list = [int(x) for x in data[6]]
        _N: int = sum(natoms)
        _type: str = data[7][0].lower()
        coords: np.ndarray = np.array(data[8: 8 + _N], dtype=float)
        cart_coords = coords @ av

        # Calculate reciprocal lattice vectors
        a1, a2, a3 = av

        # Generate atom list
        _atom_list = [f'{symbol} '*num
                      for symbol, num in zip(atoms, natoms)]

        _atom_list = ' '.join(_atom_list).split()

        bv = get_reciprocal_vectors(*av)

        # calculate the volume of the unit cell
        V = np.dot(a1, np.cross(a2, a3))

        return cls(name, scale, av, _atom_list, _N, _type,
                   coords, atoms, natoms, cart_coords, bv, V)

    @classmethod
    def from_cif(cls, filename):
        '''
        Reads a .cif file and initializes it into the structure class.

        Parameters
        ----------
        filename : str

        Returns
        -------
        Structure object
        '''
        data = io.read(filename, format='cif')
        name = data.get_chemical_formula()
        scale = 1.
        av = data.cell.array
        _atom_list = data.get_chemical_symbols()
        _N = len(_atom_list)
        _type = 'direct'
        coords = data.get_scaled_positions()

        # convert atom list into atoms and natoms
        coords = np.hstack((np.array(_atom_list).reshape(-1, 1), coords))
        coords = coords[coords[:, 0].argsort()]
        _atom_list = list(coords[:, 0])
        coords = np.array(coords[:, 1:], dtype=float)
        atoms = list(set(_atom_list))
        atoms.sort()
        natoms = [_atom_list.count(x) for x in atoms]

        cart_coords = coords @ av

        # Calculate reciprocal lattice vectors
        a1, a2, a3 = av
        bv = get_reciprocal_vectors(*av)

        # calculate the volume of the unit cell
        V = np.dot(a1, np.cross(a2, a3))

        return cls(name, scale, av, _atom_list, _N, _type,
                   coords, atoms, natoms, cart_coords, bv, V)

    def make_graph(self, atom, scaler=1.1):
        self.graph = MolGraph()
        idx = self.get_index(atom)
        atom_symbol = re.split(r'(\d+)', atom)
        vector = self.get_vector(idx, atom_symbol)
        self.shift_coords(vector, basis='F')
        coords, lv, atoms = self.generate_supercell([2, 2, 2])
        vector = np.array([-.5, -.5, -.5])
        coords += vector
        coordinates = coords @ lv
        numbers = list(map(atomic_numbers.__getitem__, atoms))
        radii = list(map(covalent_radii.__getitem__, numbers))
        masses = list(map(atomic_masses.__getitem__, numbers))
        for idx, (symbol, coordinate, radius, mass) in enumerate(zip(atoms, coordinates, radii, masses)):
            self.graph.add_node(Node(coordinate, symbol, radius, idx, mass, scaler=scaler))
        self.graph.construct_graph
        self.atom_node = list(filter(lambda node: np.allclose(node.coordinate, np.array([0., 0., 0.])), self.graph.nodes))[0]


    @property
    def cell_numbers(self):
        """
        Returns a list of numbers that correspond to species in the cell
        """
        numbers = [[i + 1] * self.natoms[i] for i in range(len(self.natoms))]
        return [x for y in numbers for x in y]

    @property
    def cell(self):
        '''
        Returns the cell structure
        '''
        return (self.lattice_vectors, self.coords, self.cell_numbers)

    @property
    def spacegroup(self):
        '''
        Returns the spacegroup of the structure
        '''
        return spglib.get_spacegroup(self.cell)

    @property
    def masses(self):
        """
        Returns the masses of the atoms in the structure
        """
        return np.array(list(map(lambda x: atomic_masses[atomic_numbers[x]], self.atom_list)))

    @property
    def get_name(self):
        '''
        Getter for name
        '''
        print(f'Name: {self.name}')
        return self.name

    @get_name.setter
    def rename(self, new_name: str):
        '''
        Setter for name
        '''
        print(f'Name changed from "{self.name}" to "{new_name}"')
        self.name = new_name

    @get_name.deleter
    def delete_name(self):
        '''
        Deleter for name
        '''
        print(f'Name "{self.name}" deleted')
        del self.name

    @property
    def as_dict(self) -> dict:
        '''
        Returns a dictionary of the structure.
        '''
        return {'name': self.name,
                'scale': self.scale,
                'lattice vectors': self.lattice_vectors,
                'reciprocal lattice vectors': self.recip_vectors,
                'volume': self.V,
                'atoms': self.atoms,
                'natoms': self.natoms,
                'atom list': self.atom_list,
                'type': self._type,
                'cart coords': self.cart_coords,
                'coords': self.coords}

    def generate_supercell(self, n: list):
        '''
        Creates a supercell of size n1 x n2 x n3.

        Parameters
        ----------
        n : list
            Expansion factor for each lattice vector.

        Returns
        -------
        coords:
            Coordinates of the supercell.
        lattice_vectors:
            Lattice vectors of the supercell.
        atom_list:
            List of atoms in the supercell.
        '''
        coords = self.coords
        av = self.lattice_vectors
        atom_list = self.atom_list

        def normalise_coord(coord, n):
            return np.array([coord[0] / n[0], coord[1] / n[1], coord[2] / n[2]])

        _n1 = list(range(n[0]))
        _n2 = list(range(n[1]))
        _n3 = list(range(n[2]))
        _n = np.array([[_n1[i], _n2[j], _n3[k]]
                       for i in range(len(_n1))
                       for j in range(len(_n2))
                       for k in range(len(_n3))])

        coords = np.array(list(map(lambda n: list(map(lambda coord: coord + n, self.coords)), _n)), dtype=float).reshape(-1, 3)
        coords = np.array(list(map(lambda coord: normalise_coord(coord, n), coords)))
        av *= n
        atom_list = np.array([atom_list] * np.prod(n)).flatten()
        return coords, av, atom_list

    def reorder_supercell(self, coords, atom_list):
        """
        Reorders the supercell to be written in POSCAR form
        """
        coords = np.concatenate((atom_list.reshape(-1, 1), coords), axis=1)
        coords = coords[coords[:, 0].argsort()]
        atom_list = coords[:, 0]
        coords = coords[:, 1:]

        atoms = []
        for i in atom_list:
            if i not in atoms:
                atoms.append(i)
        natoms = [list(atom_list).count(i) for i in atoms]
        return coords, atom_list, atoms, natoms

    def shift_coords(self, vector: np.ndarray, basis):
        """
        Shifts the structure by a constant vector.

        Parameters
        ----------
        vector : np.ndarray
            Vector to shift the structure by.
        basis : str
            Basis to shift the structure in. Either 'cart' or 'frac'.

        Returns
        -------
        None
        """
        if basis == 'C':
            self.cart_coords += vector

        if basis == 'F':
            coords = self.coords + vector
            self.coords = coords % 1

    def calculate_mesh(self) -> np.ndarray:
        '''
        Calculates K-point mesh.
        '''
        kpoint_mesh = []
        kspacing_min, kspacing_max = 0.05, 0.5
        av = self.lattice_vectors
        bv = self.recip_vectors
        bv_norm = np.array([np.linalg.norm(x) for x in bv], dtype=float)

        temp = [(i, norm) for i, norm in enumerate(bv_norm)]
        temp.sort(key=lambda x: x[1], reverse=True)

        i1, i2, i3 = [i for i, _ in temp]

        # Calculate the number of subdivisions N1, N2, N3 in the reciprocal lattice vectors 
        N_max = max(1, int(np.ceil(bv_norm[i1] / kspacing_min)))
        N_min = max(1, int(np.ceil(bv_norm[i1] / kspacing_max)))

        for n1 in range(N_min, N_max):
            min_spacing = bv_norm[i1] / n1
            if np.fabs(bv_norm[i2] - bv_norm[i1]) < 1e-5:
                n2 = n1
            else:
                n2 = int(np.ceil(bv_norm[i2] / min_spacing))
                n2 = max(n2, 1)

            if np.fabs(bv_norm[i3] - bv_norm[i2]) < 1e-5:
                n3 = n2
            else:
                n3 = int(np.ceil(bv_norm[i3] / min_spacing))
                n3 = max(n3, 1)

            if bv_norm[i2] / n2 < kspacing_max and bv_norm[i3] / n3 < kspacing_max:
                mesh = np.array([None, None, None])

            mesh[i1], mesh[i2], mesh[i3] = n1, n2, n3 
            kpoint_mesh.append(mesh)

        # calculate kpoint density
        volume = np.linalg.det(bv)
        density = np.array([[np.prod(mesh) / volume] for mesh in kpoint_mesh], dtype=float)

        return np.array(kpoint_mesh, dtype=int), density

    def calculate_encut(self, max_encut, min_encut) -> np.ndarray:
        '''
        Calculates possible ENCUT values to test for convergence.
        '''
        # check if POTCAR file exits
        if not os.path.isfile('POTCAR'):
            raise Warning('POTCAR file not found. Please generate POTCAR file first.')

        min_encut = round(min_encut / 50) * 50

        def generate_enmax(lines: list):
            for line in lines:
                if 'ENMAX' in line:
                    yield float(line.split('=')[1].split(';')[0])

        with open('POTCAR', 'r') as f:
            lines = f.readlines()
        enmax = generate_enmax(lines)
        encut = max(list(enmax)) * 1.3
        encut = round(encut / 50) * 50
        if min_encut < encut:
            print(f'Recommended ENCUT is greater than {min_encut}.')
            print(f'Using ENCUT = {encut} eV.')
        if min_encut > encut:
            encut = min_encut
        encut_list = [x for x in range(encut, max_encut, 50)]
        self.encut = encut_list
        return np.array(encut_list)

    def get_index(self, atom: str):
        '''
        finds the index of the supplied atom in the coords
        and then returns the index of the atom in the atom_list

        Parameters
        ----------
        atom : str
            atom to find the index of

        Returns
        -------
        index : int
        '''
        atom = re.split(r'(\d+)', atom)[0]
        for ind, symb in enumerate(self.atoms):
            if symb == atom:
                return ind
        return None

    def get_vector(self, idx: int, atom: str):
        """
        Returns the vector that shifts the atom to the origin.

        Parameters
        ----------
        idx : int
            Index of the atom in the coords.
        atom : str
            Atom to shift to the origin.

        Returns
        -------
        vector : np.ndarray
        """
        print(idx, atom)
        _prev = np.sum(self.natoms[:idx])
        location = int(_prev) + int(atom[1]) - 1
        return - self.coords[location]

    def get_kpath(self):
        """
        Returns the k-path for band_structure calcualtions
        """
        kpath = seekpath.get_path(self.cell, with_time_reversal=True)

        return [[x[0], kpath['point_coords'][x[0]]] for x in kpath['path']]

    def reduced_cell(self, primitive=True, niggli=False, delaunay=False, refine=False):
        """
        returns the reduced cell of the structure
        """
        cell = spglib.standardize_cell(self.cell, to_primitive=True)
        if primitive:
            return cell

        elif niggli:
            lattice = spglib.niggli_reduce(cell[0])
            return (lattice, cell[1], cell[2])

        elif delaunay:
            lattice = spglib.delaunay_reduce(cell[0])
            return (lattice, cell[1], cell[2])

        elif refine:
            return spglib.refine_cell(cell)

    def write_xyz(self, filename):
        '''
        Writes an xyz file

        Parameters
        ----------
        filename : str
            Name of the file to write to.

        Returns
        -------
        None
        '''
        coords = self.cart_coords
        atom_list = self.atom_list

        with open(filename, 'w') as f:
            f.write(f'{len(coords)}\n\n')
            for i in range(len(coords)):
                f.write('{}\t{:.10f}\t{:.10f}\t{:.10f}\n'.format(
                    atom_list[i], coords[i, 0], coords[i, 1], coords[i, 2]))

    def write_poscar(self, filename):
        '''
        Writes POSCAR file.

        Parameters
        ----------
        filename : str

        Returns
        -------
        None
        '''
        self.coords = np.array(self.coords, dtype=float)
        with open(filename, 'w') as f:
            f.write('{}\n'.format(''.join(self.name)))
            f.write('  {}\n'.format(self.scale))
            f.write('\t{:.10f}\t{:.10f}\t{:.10f}\n'.format(*self.lattice_vectors[0]))
            f.write('\t{:.10f}\t{:.10f}\t{:.10f}\n'.format(*self.lattice_vectors[1]))
            f.write('\t{:.10f}\t{:.10f}\t{:.10f}\n'.format(*self.lattice_vectors[2]))
            f.write('  {}\n'.format(' '.join(self.atoms)))
            f.write('   {}\n'.format(' '.join([str(x) for x in self.natoms])))
            f.write('{}\n'.format(self._type))
            for x in self.coords:
                f.write('  {:.10f}\t{:.10f}\t{:.10f}\n'.format(*x))


class Dope(Structure):
    """
    Class for doping structures.

    Parameters
    ----------
    Structure : Structure
    """

    def __init__(self, filename: str, dopant: str, replace: str, instances: int):
        '''
        Initializes the DOPE class

        Parameters
        ----------
        filename : str
            Name of the file to read from.
        dopant : str
            Name of the dopant to add.
        replace : str
            Name of the atom to replace.
        instances : int
            Number of instances of the dopant to add.

        Returns
        -------
        None
        '''
        extension = get_extension(filename)
        if extension == 'cif':
            cif = Structure.from_cif(filename)
            for atrr in cif.__dict__:
                setattr(self, atrr, getattr(cif, atrr))
        elif extension == 'vasp':
            poscar = Structure.from_poscar(filename)
            for atrr in poscar.__dict__:
                setattr(self, atrr, getattr(poscar, atrr))
        else:
            raise ValueError(f'Extension {extension} not found.')

        self.dopant = dopant
        self.replace = replace
        self.instances = instances

    def translate_coords(self, coords):
        """
        Translates the coordinates to be origin centered
        """
        shift = np.mean(np.array(coords, dtype=float), axis=0)
        return np.array(coords, dtype=float) - shift

    def dopant_idx(self, atom_list):
        for idx, atom in enumerate(atom_list):
            if atom == self.dopant:
                yield idx

    def Symbol_coords(self) -> np.ndarray:
        '''
        Returns the symbol and coordinates of the atoms in the structure.

        Returns
        -------
        coords : np.ndarray
        '''
        return np.hstack((np.array(self.atom_list).reshape(-1, 1), self.coords))

    def _replace_atom(self, atoms, idx=None) -> np.ndarray:
        '''
        Replaces the atom in the structure with the dopant in all possible locations.

        Parameters
        ----------
        coords : np.ndarray

        Returns
        -------
        coords : np.ndarray
        '''
        atom_lists = []
        replace = self.replace
        tmp = atoms.copy()
        for i in range(len(atoms)):
            if tmp[i] == replace and i != idx:
                tmp[i] = self.dopant
                idx = i
                atom_lists.append(tmp)
                tmp = atoms.copy()

        return np.array(atom_lists)

    @timer
    def generate_structures(self) -> np.ndarray:
        '''
        Generate the doped structures with a given number of 
        doped sites.

        Returns
        -------
        structures : np.ndarray
        '''
        # check if dopant is already in structure
        if self.dopant in self.atoms:
            atom_list = self.atom_list
            atom_list = list(map(lambda x: x.replace(self.dopant, "TEMP"), atom_list))
        else:
            atom_list = self.atom_list

        structures = self._replace_atom(atom_list)
        # transform structures into one array of structures
        instances = self.instances

        if instances == 1:
            pass
        else:
            for _ in range(instances-1):
                for i in range(len(structures)):
                    _temp = self._replace_atom(structures[i])
                    structures = np.vstack((structures, _temp))
                    structures = np.unique(structures, axis=0)

        # remove duplicarte structures
        structures = np.array([x for x in structures
                               if np.count_nonzero(
                                   x == self.dopant) == instances])

        structures = np.unique(structures, axis=0)

        structures = np.array(list(map(lambda x: list(map(lambda y: y.replace("TEMP", self.dopant), x)), structures)))

        # add coordinates to structures
        structures = np.array(
                list(map(lambda x: np.hstack((x.reshape(-1, 1), self.coords)), structures))
                )

        self._structures = structures
        print(f'Number of structures found = {len(structures)}')

    @timer
    def symmetrize(self):
        """
        Check if all structures are symmetrically unique
        """

        # list of all found structures and their atoms
        atoms_list = list(map(lambda x: x[:, 0], self._structures))
        structures = list(map(lambda x: x[:, 1:], self._structures))
        structures = list(map(lambda x: self.translate_coords(x), structures))

        reflections = np.array([
            [-1, 1, 1],
            [1, -1, 1],
            [1, 1, -1],
            [-1, -1, 1],
            [-1, 1, -1],
            [1, -1, -1],
            [-1, -1, -1]
            ])

        sym_eqiv = []
        for i in range(len(structures)):
            dopant_i = list(map(lambda x: list(structures[i][int(x)]), list((self.dopant_idx(atoms_list[i])))))
            for j in range(len(structures)):
                if i != j:
                    if not np.allclose(structures[i], structures[j]):
                        raise ValueError('Error in centering poscars')
                    dopant_j = list(map(lambda x: list(structures[j][int(x)]), list((self.dopant_idx(atoms_list[j])))))

                    for refl in reflections:
                        tmp = []
                        for di in dopant_i:
                            for dj in dopant_j:
                                if np.allclose(di, dj * refl):
                                    tmp.append(j)
                        if len(tmp) >= len(dopant_i) and j > i:
                            sym_eqiv.append([j])

        sym_eqiv = np.unique(sym_eqiv)
        print('Number of symmetrically equivalent structures = {}'.format(len(sym_eqiv)))
        if len(sym_eqiv) > 0:
            structures = np.delete(self._structures, sym_eqiv, axis=0)
        else:
            structures = self._structures
        print('Number of symmetrically unique structures = {}'.format(len(structures)))
        for i in range(len(structures)):
            structures[i] = structures[i][structures[i][:, 0].argsort()]
        self._structures = structures

    def Create_defect(self):
        '''
        Creates defects in the stucture

        Returns
        -------
        None
        '''
        structures = []
        for structure in self._structures:
            for i in range(len(structure)):
                try:
                    if structure[i][0] == self.dopant:
                        structure = np.delete(structure, i, axis=0)
                        structures.append(structure)
                except:
                    pass

        self._structures = np.array(structures)

    def write_poscars(self) -> None:
        '''
        Writes the doped structures to POSCAR files.
        '''
        structures = self._structures

        for i in range(len(structures)):
            atom_list = structures[i][:, 0]
            self.atoms, counts = np.unique(atom_list, return_counts=True)
            self.natoms = np.array([str(x) for x in counts])
            self.coords = structures[i][:, 1:]
            self.write_poscar('POSCAR-{}'.format(i+1))


class Molecule(Structure):
    '''
    A class for creating asymmetric units from a molecular crystal POSCAR file.

    Parameters
    ----------
    Stucture : Structure
    '''

    def __init__(self,
                 filename: str,
                 atom: str,
                 scaler: float,
                 ):
        '''
        Initializes the Asymmetric_unit class.
        '''
        extension = get_extension(filename)
        if extension == 'cif':
            cif = Structure.from_cif(filename)
            for atrr in cif.__dict__:
                setattr(self, atrr, getattr(cif, atrr))
        elif extension == 'vasp':
            poscar = Structure.from_poscar(filename)
            for atrr in poscar.__dict__:
                setattr(self, atrr, getattr(poscar, atrr))
        else:
            raise ValueError(f'Extension {extension} not found.')
        self.atom = atom
        self.scaler = scaler

    def translate(self):
        '''
        Performs a translation and expansion to ensure a
        whole molecule is included.
        '''
        idx = self.get_index(self.atom)
        atom_symb = re.split(r'(\d+)', self.atom)
        shift_vector = self.get_vector(idx, atom_symb)
        self.shift_coords(shift_vector, basis='F')
        coords, lattice_vectors, atom_list = self.generate_supercell([2, 2, 2])
        shift_vector = np.array([-.5, -.5, -.5])
        coords += shift_vector
        cart_coords = coords @ lattice_vectors
        return cart_coords, atom_list

    def origin_index(self, coords) -> int:
        '''
        Finds the index of the origin atom.
        '''
        for ind, val in enumerate(coords):
            if np.allclose(val, [0, 0, 0]):
                return ind

    def nearest_neighbours(self, coords: np.ndarray, point: np.ndarray,
                            bond_max: float) -> np.ndarray:
        """
        Uses a KDTree to find the nearest neighbours of a point.

        Parameters
        ----------
        coords : np.ndarray
            The coordinates of the atoms.
        point : np.ndarray
            The central index coordinates.
        bond_max : float
            The maximum bond length.

        Returns
        -------
        coords : np.ndarray
            The coordinates of the nearest neighbours.
        """
        neighbours = sp.spatial.KDTree(coords[:, 1:], leafsize=10,
                                       compact_nodes=True, balanced_tree=True)
        dist, ind = neighbours.query(point, k=10)
        ind = [ind[i] for i in range(len(ind)) if dist[i] < bond_max]
        coords = np.array([coords[i] for i in ind])
        return coords

    @timer
    def find_molecule(self) -> None:
        '''
        Finds a molecule within a molecular crsytal structure.
        '''
        self.make_graph(self.atom, self.scaler)
        mol_graph = self.graph.extract_fragment(self.atom_node)
        mol_graph.plot
        mol_graph.to_xyz('{}.xyz'.format("".join(mol_graph.graph_formula)), mass_centered=True)

    def get_neighbours(self, sphere_coords, total_coords, atom):
        """
        Returns an array of the atom and its neighbours
        """
        sphere_coords = np.concatenate(
                (sphere_coords,
                 self.nearest_neighbours(
                     total_coords, atom[1:], self.bond_max
                     )
                 ), axis=0)
        return np.unique(sphere_coords, axis=0)

    @timer
    def get_coordination_sphere(self, coordination_sphere: int):
        """
        Finds the specified coordination sphere of a contiuous crystaline
        structure.

        Parameters
        ----------
        coordination_sphere : int
            The coordination sphere to find.
        """
        self.make_graph(self.atom, self.scaler)
        self.qm_graph = self.graph.coordination_sphere(self.atom_node, coordination_sphere)

    def bridge_dimer(self):
        self.make_graph(self.atom, self.scaler)
        nodes = self.graph.nodes
        nodes = list(filter(lambda x: x != self.atom_node, nodes))
        nodes = list(filter(lambda x: x.symbol == self.atom_node.symbol, nodes))
        distances = list(map(lambda node: node - self.atom_node, nodes))
        min_node = nodes[np.argmin(distances)]
        self.graph.plot

        qm_graph = self.graph.bridge_nodes(self.atom_node, min_node)
        self.qm_graph = qm_graph


def get_extension(filename):
    """
    Gets the extension of a file.

    Parameters
    ----------
    filename : str
        The filename.

    Returns
    -------
    file_type: str
        The file extension.
    """
    parseable = ['cif', 'vasp', 'h5', 'hdf5']
    try:
        file_type = filename.split('.')[-1]
        if file_type not in parseable:
            raise ValueError('File type {} not supported.'.format(file_type))
        return file_type

    except:
        return 'vasp'


def get_reciprocal_vectors(a1, a2, a3):
    b1 = 2 * np.pi * np.cross(a2, a3) / (a1@np.cross(a2, a3))
    b2 = 2 * np.pi * np.cross(a3, a1) / (a2@np.cross(a3, a1))
    b3 = 2 * np.pi * np.cross(a1, a2) / (a3@np.cross(a1, a2))
    return np.array([b1, b2, b3])



# Imports
import socket


class SubmissionWriter():
    def __init__(self, title, cores, vasp_type):
        self.title = title
        self.cores = cores
        self.vasp_type = vasp_type

    @property
    def hostname(self):
        hostname = socket.gethostname()
        if 'csf3' in hostname:
            return 'csf3'
        elif 'csf4' in hostname:
            return 'csf4'
        else:
            return 'csf3'

    @property
    def handle_cores_csf3(self):
        if self.cores < 32:
            return '#$ -pe smp.pe {}'.format(self.cores)
        else:
            return '#$ -pe mpi-24-ib.pe {}'.format(self.cores)

    @property
    def handle_nodes_csf4(self):
        if self.cores < 40:
            return '#SBATCH -p multicore'
        else:
            return '#SBATCH -p multinode'

    @property
    def handle_vasp_type(self):
        options = {
                'standard': 'vasp_std',
                'gamma': 'vasp_gam',
        }

        if self.hostname == 'csf3':
            return 'mpirun -np {} {}_par'.format(
                    self.cores, options[self.vasp_type])
        if self.hostname == 'csf4':
            return 'mpirun -np {} {}'.format(
                    self.cores, options[self.vasp_type])

    def write_csf3_header(self, FileObject):
        FileObject.write("#!/bin/bash --login\n")
        FileObject.write("#$ -cwd\n")
        FileObject.write('{}\n'.format(self.handle_cores_csf3))
        FileObject.write("#$ -N {}\n".format(self.title))

    def write_csf4_header(self, FileObject):
        FileObject.write("#!/bin/bash --login\n")
        FileObject.write("{}\n".format(self.handle_nodes_csf4))
        FileObject.write("#SBATCH -n {}\n".format(self.cores))
        FileObject.write("#SBATCH --job-name={}\n".format(self.title))

    def write_module_load(self, FileObject):
        if self.hostname == 'csf3':
            FileObject.write('module load apps/intel-19.1/vasp/5.4.4\n')
        elif self.hostname == 'csf4':
            FileObject.write('module load vasp/5.4.4-iomkl-2020.02\n')

    def write_body(self, FileObject):
        FileObject.write('echo "Started running at $(date)"\n')
        FileObject.write('{}\n'.format(self.handle_vasp_type))
        FileObject.write('echo "Finished running at $(date)"\n')

    def write_array_header_csf3(self, array_file, FileObject):
        with open(array_file, 'r') as f:
            t_max = len([line for line in f.readlines()])
        FileObject.write('#$ -t 1-{}\n'.format(t_max))
        FileObject.write('module load apps/intel-19.1/vasp/5.4.4\n\n')
        FileObject.write('TASK_ID=$SGE_TASK_ID\n')
        FileObject.write('STEM=$(sed -n "${{TASK_ID}}p" {})\n'.format(array_file))
        FileObject.write('export CurrDir=$(pwd -P)\n')

    def write_array_header_csf4(self, array_file, FileObject):
        with open(array_file, 'r') as f:
            t_max = len([line for line in f.readlines()])
        FileObject.write('#SBATCH -a 1-{}\n'.format(t_max))
        FileObject.write('module load vasp/5.4.4-iomkl-2020.02\n')
        FileObject.write('TASK_ID=$SLURM_ARRAY_TASK_ID\n')
        FileObject.write('STEM=$(sed -n "${{TASK_ID}}p" {})\n'.format(array_file))
        FileObject.write('export CurrDir=$(pwd -P)\n')

    def write_phonopy_body(self, FileObject):
        FileObject.write('echo "Started running at $(date)"\n')
        FileObject.write('if [ ! -d "$CurrDir/$STEM" ]; then\n')
        FileObject.write('\tmkdir $CurrDir/$STEM\n')
        FileObject.write('\tcd $CurrDir\n')
        FileObject.write('fi\n\n')
        FileObject.write('mv $CurrDir/POSCAR-$STEM $CurrDir/$STEM/\n')
        FileObject.write('cp $CurrDir/INCAR $CurrDir/$STEM/\n')
        FileObject.write('cp $CurrDir/KPOINTS $CurrDir/$STEM/\n')
        FileObject.write('cp $CurrDir/POTCAR $CurrDir/$STEM/\n')
        FileObject.write('cd $CurrDir/$STEM/\n')
        FileObject.write('mv POSCAR-$STEM POSCAR\n')
        FileObject.write('{}\n'.format(self.handle_vasp_type))
        FileObject.write('cd $CurrDir\n\n')
        FileObject.write('echo "Finished running at $(date)"\n')

    def write_convergence_body(self, FileObject):
        FileObject.write('echo "Started running at $(date)"\n')
        FileObject.write('cd $CurrDir/$STEM\n')
        FileObject.write('{}\n'.format(self.handle_vasp_type))
        FileObject.write('cd $CurrDir\n\n')
        FileObject.write('echo "Finished running at $(date)"\n')



"""
Convergence testing for vasp calculations
"""
import os
import shutil
import csv
import numpy as np

from .structure import Structure
from .parser import ParseOUTCAR, ParseOSZICAR
from .core import generate_kpoints

import matplotlib.pyplot as plt


def make_dirs(combinations):
    """
    Make directories for each combination of parameters
    and copy the input files into them
    """
    cwd = os.getcwd()

    for c in combinations:
        dir = '{}{}{}_{}'.format(c[0][0], c[0][1], c[0][2], c[1])
        if not os.path.exists(dir):
            os.makedirs(dir)
        shutil.copy(os.path.join(cwd, 'INCAR'), os.path.join(cwd, dir))
        shutil.copy(os.path.join(cwd, 'POTCAR'), os.path.join(cwd, dir))
        shutil.copy(os.path.join(cwd, 'POSCAR'), os.path.join(cwd, dir))
        os.chdir(dir)
        generate_kpoints(c[0])
        with open('INCAR', 'a') as f:
            f.write('ENCUT = {}\n'.format(c[1]))
        os.chdir(cwd)

    with open('convergence.txt', 'w') as f:
        for i in combinations:
            f.write('{}{}{}_{}\n'.format(i[0][0], i[0][1], i[0][2], i[1]))


def get_combinations(filename, max_encut, min_encut):
    """
    Convergence test calculation set up
    """

    if os.path.splitext(filename)[1] == ".cif":
        filename = 'POSCAR'

    poscar = Structure.from_poscar(filename)

    # get the kpoint mesh and densities
    mesh, density = poscar.calculate_mesh()

    encut_values = poscar.calculate_encut(max_encut, min_encut)

    combinations = [[x, y, z] for x, z in zip(mesh[:5], density[:5])
                    for y in encut_values]

    return combinations


def extract_data(combinations):
    """
    extracting data from the calculations
    for post processing
    """

    with open('conv.csv', 'w') as output_writer:
        output_writer_csv = csv.writer(output_writer, delimiter=',')

        output_writer_csv.writerow(
                ["Cutoff [eV]", "nk_1", "nk_2", "nk_3", "nk,red",
                 r"\rho_k [A^-3]", "E_0 [eV]", "p_ext [kbar]", "#SCF", "t [s]"]
                )

    for c in combinations:
        dir = '{}{}{}_{}'.format(c[0][0], c[0][1], c[0][2], c[1])
        outcar = ParseOUTCAR(os.path.join(dir, 'OUTCAR'))
        oszicar = ParseOSZICAR(os.path.join(dir, 'OSZICAR'))

        energy = list(oszicar.energy)
        ext_pressure = list(outcar.external_pressure)
        num_scf = list(oszicar.electronic_steps)
        time = list(outcar.elapsed_time)
        n_k_red = np.prod(c[0])

        with open('conv.csv', 'a') as output_writer:
            output_writer_csv = csv.writer(output_writer, delimiter=',')
            output_writer_csv.writerow(
                    [c[1], c[0][0], c[0][1], c[0][2], str(n_k_red), c[2][0],
                     str(energy[0]), str(ext_pressure[0]),
                     str(num_scf[0]), str(time[0])]
                    )


def kpoint_plot(data, num_atoms, plot_gradient=False):
    unique_encut = np.unique(data[r"Cutoff [eV]"])
    height = int(np.ceil(len(unique_encut) / 2))
    width = 2

    fig, axs = plt.subplots(height, width, figsize=(10, height * 5))

    for i, encut in enumerate(unique_encut):
        encut_data = data[data[r"Cutoff [eV]"] == encut]
        energy_per_atom = encut_data["E_0 [eV]"] / num_atoms
        gradient = np.gradient(energy_per_atom, encut_data[r"\rho_k [A^-3]"])

        axs[int(i // width), int(i % width)].plot(
                encut_data[r"\rho_k [A^-3]"],
                energy_per_atom,
                "k",
                label=r"ENCUT = {}".format(encut),
                marker="s"
                )
        axs[int(i // width), int(i % width)].set_xlabel(r"$\rho_k$ (Å$^{-3}$)")
        axs[int(i // width), int(i % width)].set_ylabel("Energy per atom (eV)")
        axs[int(i // width), int(i % width)].set_title("ENCUT = {}".format(encut))

        if plot_gradient:
            grad_ax = axs[int(i // width), int(i % width)].twinx()
            grad_ax.plot(
                    encut_data[r"\rho_k [A^-3]"],
                    gradient,
                    "--r",
                    marker="s"
                    )
            grad_ax.set_ylabel("Gradient (eV/Å$^{-3}$)")
            grad_ax.yaxis.label.set_color("red")

    fig.tight_layout()
    fig.savefig("kpoint_convergence.pdf", dpi=300)


def cutoff_plot(data, num_atoms, plot_gradient=False):
    unique_rho = np.unique(data[r"\rho_k [A^-3]"])
    height = int(np.ceil(len(unique_rho) / 2))
    width = 2

    fig, axs = plt.subplots(height, width, figsize=(10, height * 5))

    for i, rho in enumerate(unique_rho):
        rho_data = data[data[r"\rho_k [A^-3]"] == rho]
        energy_per_atom = rho_data["E_0 [eV]"] / num_atoms
        gradient = np.gradient(energy_per_atom, rho_data[r"Cutoff [eV]"])
        double_gradient = np.gradient(gradient, rho_data[r"Cutoff [eV]"])

        axs[int(i // width), int(i % width)].plot(
                rho_data["Cutoff [eV]"],
                energy_per_atom,
                "k",
                label=r"$\rho_k$ = {}".format(rho),
                marker="s"
                )
        axs[int(i // width), int(i % width)].set_xlabel("ENCUT (eV)")
        axs[int(i // width), int(i % width)].set_ylabel("Energy per atom (eV)")
        axs[int(i // width), int(i % width)].set_title(r"$\rho_k$ = {}".format(rho))

        if plot_gradient:
            grad_ax = axs[int(i // width), int(i % width)].twinx()
            grad_ax.plot(
                    rho_data["Cutoff [eV]"],
                    gradient,
                    "--r",
                    label=r"Gradient",
                    marker="o"
                    )
            grad_ax.set_ylabel("Gradient")
            grad_ax.yaxis.label.set_color("red")

            double_grad_ax = axs[int(i // width), int(i % width)].twinx()
            double_grad_ax.plot(
                    rho_data["Cutoff [eV]"],
                    double_gradient,
                    "--b",
                    label=r"Double derivative",
                    marker="o"
                    )
            double_grad_ax.set_ylabel("Double derivative")


    fig.tight_layout()
    fig.savefig("encut_convergence.pdf", dpi=300)



"""
Handles the automation of phonopy calculations.
"""
import os


def store_disp_files():
    """
    Identifies the number of 'POSCAR-XXX' files in the current directory.
    """
    disp_files = [x for x in os.listdir()
                  if x.startswith('POSCAR-')]

    disp_files = [x.split('-')[1] for x in disp_files]
    print('Found {} displacement files'.format(len(disp_files)))
    disp_files = sorted(disp_files, key=int)

    with open('phonon_files.txt', 'w') as f:
        for i in disp_files:
            f.write('{}\n'.format(str(i)))


def check_config():
    """
    Checks if the config file exists.
    """
    if os.path.isfile(os.path.join(os.path.expanduser('~'),
                                   '.vasp_suite_configs/phonon.ini')):
        return True
    else:
        return False
